/**
  *  \file game/cargotr.h
  *  \brief Cargo Transfer
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_CARGOTR_H
#define PCC_V2_GAME_CARGOTR_H

#include "game/cargotype.h"
#include "game/transact.h"
#include "cpluslib/sigslot.h"
#include "cpluslib/ptrvector.h"
#include "game/cargo.h"

/*! \class GCargoTransfer
    \brief Cargo Transfer

    This class provides everything necessary to transfer cargo between objects.
    It holds (and controls life of) a number of GCargoContainer objects,
    similar to a GMultiTransaction.
    It consolidates the cargo container's sig_changed event into one single event.

    There can be an arbitrary number of objects, although the most usual case is two.
    The move() method is intended for two objects, for multiple objects, use moveFT(). */
class GCargoTransfer : public GTransaction {
 public:
    GCargoTransfer();
    ~GCargoTransfer();

    // Container management
    GCargoContainer& addNewContainer(GCargoContainer* obj);
    GCargoContainer& operator[](std::size_t i);
    std::size_t size() const;

    // Cargo transfer
    int32_t move(GCargoType type, int32_t amount, std::size_t from, bool partial, bool s2m);
    int32_t moveFT(GCargoType type, int32_t amount, std::size_t from, std::size_t to, bool partial, bool s2m);
    bool    doUnload(bool s2m);

    // Status
    bool    isValid() const;
    GError  getError();

    /** Change signal. Raised whenever cargo is moved. */
    Signal0 sig_changed;

 protected:
    void cancel();
    void commit();

 private:
    ptr_vector<GCargoContainer> objs;
};

/** Get object on side.
    \param side which object you want */
inline GCargoContainer&
GCargoTransfer::operator[](std::size_t side)
{
    return *objs[side];
}

/** Get number of containers. */
inline std::size_t
GCargoTransfer::size() const
{
    return objs.size();
}
#endif
