/**
  *  \file game/mission.cc
  *  \brief Starship Mission Management
  *
  *  This module defines functions to manage ship mission definitions.
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <cctype>
#include "game/mission.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/gamedirs.h"
#include "game/host.h"
#include "game/player.h"        // getPlayerIdFromChar
#include "game/reg.h"           // allowRegistered
#include "game/ship.h"
#include "int/bytecode.h"
#include "int/exec.h"
#include "int/expr.h"           // IntCompilationContext
#include "int/if/shipif.h"
#include "int/process.h"        // runTemporaryProcess
#include "int/simple.h"         // compileExpression
#include "int/value.h"
#include "io/streams.h"
#include "io/textfile.h"
#include "util/console.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/modinit.h"
#include "util/nls.h"

/** Extract mission parameter from string.
    Updates the parameter name in-place.
    \param msn   [in/out] mission to update
    \param line  [in] mission.ini line
    \param tpos  [in] position of this parameter's token ("I:") in the line
    \param opos  [in] position of other parameter's token ("T:") in the line
    \param limit [in] position after parameters (closing paren) */
static void
extractMissionParameter(GMission& msn,
                        GMission::Parameter which,
                        const string_t& line,
                        const string_t::size_type tpos,
                        const string_t::size_type opos,
                        const string_t::size_type limit)
{
    if (tpos != line.npos) {
        string_t::size_type tend = limit-1;
        if (opos != line.npos && opos > tpos)
            tend = opos;
        while (line[tend-1] == ' ' || line[tend-1] == ',')
            --tend;
        msn.setParameterName(which, line.substr(tpos+2, tend-tpos-2));
    }
}

/******************************** GMission *******************************/

/** Construct blank mission definition. */
GMission::GMission()
    : number(0), flags(0), hotkey(0)
{
    arg_spec[0] = arg_spec[1] = 0;
}

/** Construct mission definition from mission.cc line. */
GMission::GMission(int number, string_t desc_line)
    : number(number), race_mask(GPlayerSet::allUpTo(NUM_OWNERS)), flags(0), hotkey(0)
{
    arg_spec[0] = arg_spec[1] = 0;
    parseDescription(desc_line);
}

/** Destructor. */
GMission::~GMission()
{ }

/** Get name for parameter. */
string_t
GMission::getParameterName(Parameter id) const
{
    if (arg_names[id].empty()) {
        if (id == TowParameter) {
            return "Tow";
        } else {
            return "Intercept";
        }
    }
    return arg_names[id];
}

/** True iff mission (might) work on the specified ship. Actually,
    this should return true if we want the user to be able to set the
    mission on the specified ship. */
bool
GMission::worksOn(GShip& ship) const
{
    // Don't allow mission 1 for SRace
    if (!host.isMissionAllowed(number)) {
        return false;
    }

    // Check race mask
    if (!race_mask.contains(raceMissionId(ship.getRealOwner()))) {
        return false;
    }

    // Check registration status
    if ((flags & mf_Registered) && !allowRegistered(ship.getRealOwner())) {
        return false;
    }

    // Check waypoint permission
    if ((flags & mf_Waypoint) != 0 && ship.isFleetMember()) {
        return false;
    }

    // Check mission.cc condition
    IntBCORef bco = compileExpression(exp_condition, IntCompilationContext());
    if (bco) {
        // Run it
        IntExecutionContext exec("Temporary: Mission");
        exec.pushNewContext(new IntShipContext(ship.getId()));
        exec.pushNewFrame(new IntExecutionFrame(bco));
        if (runTemporaryProcess(exec)) {
            // Evaluated successfully
            if (getBoolValue(exec.getResult()) <= 0) {
                return false;
            }
        } else {
            // Mission is not permitted if expression fails (same as in PCC 1.x)
            return false;
        }
    }

    return true;
}

/** Get label for this mission if it is set on /ship/. This name is
    intended to be displayed on the ship screen.

    NOTE: This function intentionally does not take a 'const GShip',
    since the interpreter might (need to) modify it. */
string_t
GMission::getLabel(GShip& ship) const
{
    // Try to find the name using the configured expression, if any
    string_t result;
    IntBCORef bco = compileExpression(exp_label, IntCompilationContext());
    if (bco) {
        // Run it
        IntExecutionContext exec("Temporary: Mission");
        exec.pushNewContext(new IntShipContext(ship.getId()));
        exec.pushNewFrame(new IntExecutionFrame(bco));
        if (runTemporaryProcess(exec)) {
            // Check result
            result = IntValue::toStringOf(exec.getResult(), false);
        } else {
            // Ignore script errors
            // FIXME: NO! we should not ignore execution errors. At least, PCC1.x doesn't.
        }
    }

    // Use default
    if (result.size() == 0)
        result = name;

    return result;
}

/** Check for warning. This checks the "warning" expression or,
    if that's not present, the regular condition.

    NOTE: This function intentionally does not take a 'const GShip',
    since the interpreter might (need to) modify it. */
bool
GMission::isWarning(GShip& ship) const
{
    // Try to find the expression
    IntBCORef bco;
    if (!exp_warning.empty()) {
        bco = compileExpression(exp_warning, IntCompilationContext());
    } else if (!exp_condition.empty()) {
        bco = compileExpression(exp_condition, IntCompilationContext());
    } else {
        // leave it 0
    }

    // Try to run it
    bool result = false;
    if (bco) {
        IntExecutionContext exec("Temporary: Mission");
        exec.pushNewContext(new IntShipContext(ship.getId()));
        exec.pushNewFrame(new IntExecutionFrame(bco));
        if (runTemporaryProcess(exec)) {
            // Check result
            result = getBoolValue(exec.getResult()) <= 0;
        } else {
            // A script error means there is a warning
            result = true;
        }
    }
    return result;
}

/** Parse mission.cc description line. This will set all affected
    members according to that definition.

    \param desc_line line from mission.cc, sans mission number
    (i.e. "!h#,Tow a ship") */
void
GMission::parseDescription(const string_t& desc_line)
{
    string_t::size_type n = 0;

    int arg_type = mat_Integer, arg_flags = 0;
    while (n < desc_line.length() && desc_line[n] != ',') {
        switch (desc_line[n++]) {
         case '+': race_mask = parsePlayerList(desc_line, n); break;
         case '-': race_mask = parsePlayerList(desc_line, n) ^ GPlayerSet::allUpTo(NUM_OWNERS); break;
         case 'r': flags |= mf_Registered; break;
         case 'i': flags |= mf_Waypoint; break;
         case 'o': arg_flags |= ma_Own; break;
         case '!': arg_flags |= ma_NotThis; break;
         case 'p': arg_type = mat_PID; break;
         case 's': arg_type = mat_SID; break;
         case 'h': arg_type = mat_SID_here; break;
         case 'b': arg_type = mat_BID; break;
         case 'y': arg_type = mat_Player; break;
         case '*':
            arg_spec[0] = arg_type | arg_flags;
            break;
         case '#':
            arg_spec[1] = arg_type | arg_flags;
            break;
        }
    }
    if (n < desc_line.length()) {
        name = desc_line.substr(n+1);

        /* assign hot-key if so desired */
        string_t::size_type p = name.find('~');
        if (p != name.npos && p+1 < name.length()) {
            /* "~x" assigns a hot-key. Only standard keys so far. */
            using namespace std;
            uint8_t key = tolower((unsigned char) name[p+1]);
            if (key > ' ' && key < 127)
                hotkey = key;
            name.erase(p, 1);
        }

        /* assign standard short-name */
        short_name.assign(name, 0, 7);
    }
}

/****************************** GMissionList *****************************/

/// Global instance of the mission table.
GMissionList GMissionList::instance;

/** Construct empty mission list. */
GMissionList::GMissionList()
    : used_letters(0)
{ }

/** Construct mission list for a ship. The list will contain all missions
    doable by the specified ship.
    \param other complete mission definition
    \param ship  ship for which to prepare the list */
GMissionList::GMissionList(const GMissionList& other, GShip& ship)
    : used_letters(0)
{
    for (iterator i = other.begin(); i != other.end(); ++i)
        if (i->worksOn(ship))
            data.push_back(*i);
}

/** Destructor. */
GMissionList::~GMissionList()
{ }

/** Add mission to list. This will add a (copy of) /msn/ to the
    mission list. If the mission was already defined, the call is
    ignored.
    \return true iff mission was added, false if call was ignored */
bool
GMissionList::addMission(const GMission& msn)
{
    /* do not add if it's already there. A mission is considered
       already there if its race mask is not disjoint to ours. (This
       might sound counter-intuitive, but it's used to refuse
       replacing a MISSION.CC definition by a MISSION.INI one;
       MISSION.CC generally has better race masks and checking for
       subsets would make MISSION.INI always override those.) */
    for (iterator i = begin(); i != end(); ++i) {
        if (i->getNumber() != msn.getNumber())
            continue;
        if (!msn.getRaceMask().containsAnyOf(i->getRaceMask()))
            continue;
        return false;
    }

    /* assign hot-key. Do this without modifying /msn/. */
    char c = msn.getHotkey();
    if (c == 0) {
        if (msn.getNumber() >= 0 && msn.getNumber() < 10) {
            c = '0' + msn.getNumber();
        } else {
            for (int i = 0; i < 26; ++i) {
                if (!(used_letters & (1 << i))) {
                    c = 'a' + i;
                    break;
                }
            }
            if (c == 0) {
                /* all letters used up */
                used_letters = 0;
                c = 'a';
            }
        }
    }
    ASSERT(c);
    if (c >= 'a' && c <= 'z')
        used_letters |= 1 << (c - 'a');

    data.push_back(msn);
    data.back().setHotkey(c);
    return true;
}

static bool
missionCompare(const GMission& a, const GMission& b)
{
    if (a.getNumber() < b.getNumber())
        return true;
    if (a.getNumber() > b.getNumber())
        return false;
    return a.getRaceMask().toInteger() < b.getRaceMask().toInteger();
}

/** Sort mission list. Brings list into numerical mission order, which
    will then be used on the mission screen. */
void
GMissionList::sort()
{
    std::sort(data.begin(), data.end(), missionCompare);
}

/** Clear this mission list instance. */
void
GMissionList::clear()
{
    used_letters = 0;
    data.clear();
}

/** Get mission by number/player.
    \param id mission number
    \param race_mask races (not players!) for which you want the mission. */
const GMission*
GMissionList::getMissionByNumber(int id, GPlayerSet race_mask) const
{
    for (iterator i = begin(); i != end(); ++i)
        if (i->getNumber() == id && i->getRaceMask().containsAnyOf(race_mask))
            return &*i;
    return 0;
}

/** Get mission by ship.
    \param ship Ship whose mission to query */
const GMission*
GMissionList::getMissionByShip(const GShip& ship)
{
    return getMissionByNumber(ship.getMission(), GPlayerSet(raceMissionId(ship.getRealOwner())));
}

/** Load mission.cc file.
    \param in      stream */
void
GMissionList::loadFromFile(Stream& in)
{
    TextFile tf(in);
    string_t line;
    bool have_mission = false;

    while (tf.getLine(line)) {
        line = strTrim(line);
        if (line.empty() || line[0] == ';')
            continue;

        string_t::size_type p = line.find_first_of("=,");
        if (p == string_t::npos) {
            console.write(LOG_ERROR, format(_("%s:%d: missing delimiter"), in.getName(), tf.getLineNr()));
        } else if (line[p] == ',') {
            int mnum;
            if (!stringValue(strTrim(line.substr(0, p)), mnum) || mnum < 0 || mnum > 10000) {
                console.write(LOG_ERROR, format(_("%s:%d: invalid mission number"), in.getName(), tf.getLineNr()));
                have_mission = false;
                continue;       // with next mission
            }

            line.erase(0, p+1);
            try {
                have_mission = addMission(GMission(mnum, line));
            }
            catch(UserException& e) {
                console.write(LOG_ERROR, format("%s:%d: %s", in.getName(), tf.getLineNr(), e.what()));
            }
        } else {
            /* additional assignments. */
            if (!have_mission)
                continue;

            string_t lhs = strTrim(line.substr(0, p));
            line = strTrim(line.substr(p+1));
            if (stringMatch("I", lhs)) {
                // Intercept name
                data.back().setParameterName(GMission::InterceptParameter, line);
            } else if (stringMatch("J", lhs)) {
                // Tow name
                data.back().setParameterName(GMission::TowParameter, line);
            } else if (stringMatch("Shortname", lhs)) {
                // short mission name
                data.back().setShortName(line);
            } else if (stringMatch("Condition", lhs)) {
                // condition which says whether this mission is allowed
                // to be set on this ship
                data.back().setConditionExpression(line);
            } else if (stringMatch("Text", lhs)) {
                // Text to show if mission is set
                data.back().setLabelExpression(line);
            } else if (stringMatch("Willwork", lhs)) {
                // condition which says whether the mission will work,
                // to warn the user when he sets cloaking on a damaged
                // ship.
                data.back().setWarningExpression(line);
            } else if (stringMatch("Onset", lhs)) {
                // OnSet hook: command to be called when mission is set
                data.back().setSetCommand(line);
            }
        }
    }
}

/** Load from MISSION.INI file. */
void
GMissionList::loadFromIniFile(Stream& in)
{
    TextFile tf(in);
    tf.setCharacterSet(getGameCharacterSet());
    string_t line;
    while (tf.getLine(line)) {
        line = strTrim(line);

        string_t::size_type index = line.find(' ');
        if (index == string_t::npos)
            continue;

        int number;
        if (!stringValue(line.substr(0, index), number) || number < 10)
            continue;

        /* Start building the mission */
        GMission new_mission(number, string_t());

        /* "/123" race flags */
        line.erase(0, index+1);
        index = line.rfind('/');
        if (index != line.npos) {
            GPlayerSet player_restriction;
            string_t::size_type num = 1;
            while (index+num < line.size()) {
                int pl = getPlayerIdFromChar(line[index+num]);
                if (pl > 0 && pl <= NUM_PLAYERS) {
                    player_restriction += pl;
                    ++num;
                } else {
                    break;
                }
            }
            if (player_restriction.nonempty()) {
                new_mission.setRaceMask(player_restriction);
                line.erase(index, num);
            }
        }

        /* "*#" parameter flags */
        index = line.size();
        while (index > 0) {
            if (line[index-1] == '*') {
                new_mission.setParameterType(GMission::InterceptParameter, GMission::mat_Integer);
                --index;
            } else if (line[index-1] == '#') {
                new_mission.setParameterType(GMission::TowParameter, GMission::mat_Integer);
                --index;
            } else if (line[index-1] == ' ') {
                --index;
            } else if (line[index-1] == ')') {
                /* parenthesized expression: "(I:Foo T:Bar)" */
                string_t::size_type beg = line.rfind('(');
                if (beg == line.npos)
                    break;

                /* Find "I:" and "T:" tokens */
                string_t::size_type ipos = strUCase(line).find("I:", beg);
                string_t::size_type tpos = strUCase(line).find("T:", beg);

                /* Extract "I:" and "T:" */
                extractMissionParameter(new_mission, GMission::InterceptParameter, line, ipos, tpos, index);
                extractMissionParameter(new_mission, GMission::TowParameter,       line, tpos, ipos, index);

                /* Trim whitespace and exit */
                index = beg;
                while (index > 0 && line[index-1] == ' ')
                    --index;
                break;
            } else {
                break;
            }
        }
        line.erase(index);

        /* Synthesize "active" expression. */
        if (new_mission.getParameterType(GMission::InterceptParameter) != 0 || new_mission.getParameterType(GMission::TowParameter) != 0) {
            string_t expr = quoteString(line + " (");
            if (new_mission.getParameterType(GMission::InterceptParameter)) {
                expr += "&'I:'&Mission.Intercept";
            }
            if (new_mission.getParameterType(GMission::TowParameter)) {
                if (new_mission.getParameterType(GMission::InterceptParameter))
                    expr += "&' T:'";
                else
                    expr += "&'T:'";
                expr += "&Mission.Tow";
            }
            expr += "&')'";
            new_mission.setLabelExpression(expr);
        }

        /* Add it */
        new_mission.setName(line);
        addMission(new_mission);
    }
}


/** Initialize missions. This loads the mission definition files into
    the global GMissionList instance. It is idempotent, i.e. can be
    called at any time. */
static void
initMissions()
{
    GMissionList::getInstance().clear();
    Ptr<Stream> s = spec_file_dir->openFileNT("missions.usr", Stream::C_READ);
    if (s) {
        GMissionList::getInstance().loadFromFile(*s);
    }
    s = spec_file_dir->openFileNT("mission.cc", Stream::C_READ);
    if (s) {
        GMissionList::getInstance().loadFromFile(*s);
    }
    s = spec_file_dir->openFileNT("mission.ini", Stream::C_READ);
    if (s) {
        GMissionList::getInstance().loadFromIniFile(*s);
    }

    GMissionList::getInstance().sort();
}

/** Check whether a mission causes the ship to cloak. */
bool
isMissionCloaking(int mission_id, int owner)
{
    return (mission_id == GMission::msn_Cloak
            || (host.isPHost() && mission_id == config.ExtMissionsStartAt() + GMission::pmsn_Cloak)
            || (raceMissionId(owner) == 3
                && (mission_id == GMission::msn_Special
                    || (host.isPHost()
                        && (mission_id == config.ExtMissionsStartAt() + GMission::pmsn_Special
                            || mission_id == config.ExtMissionsStartAt() + GMission::pmsn_StandardSuperSpy)))));
}

const ModuleInitializer mod_missions[] = {
    /* Load mission definitions upon entering NoGame state or a directory. */
    { OnEnterNoGame | OnEnterDirectory, initMissions },
    { 0, 0 },
};
