/**
  *  \file gfx/font.cc
  *  \brief Font Drivers
  *
  *  This module contains the user-called font output routines
  *  that support alignment and limiting.
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/font.h"
#include "cpluslib/assert.h"
#include "cpluslib/utf8.h"

/** Output Text, using Alignment Parameters.
    \param ctx    Output
    \param x,y    Anchor point
    \param text   Text to output */
void
outText(GfxContext& ctx, int x, int y, string_t text)
{
    GfxFont& fnt   = ctx.getFont();
    GfxPoint align = ctx.getTextAlign();
    if (align.x)
        x -= fnt.getTextWidth(text) * align.x / 2;
    if (align.y)
        y -= fnt.getTextHeight(text) * align.y / 2;
    fnt.outText(ctx, GfxPoint(x, y), text);
}

/** Output Text, using Alignment Parameters.
    \overload */
void
outText(GfxContext& ctx, int x, int y, const char* text)
{
    outText(ctx, x, y, string_t(text));
}

/** Output Text with fixed maximum width. Outputs the text so that it
    occupies at most maxw pixels on the screen. When non-transparent output
    is used, draws appropriate boxes so that exactly the specified area is
    covered. Otherwise like outText().

    This function moves the graphics cursor, depending on the text
    justification:
    - when using left-justified text, the cursor is set to just after
      the output text;
    - when using right-justified text, the cursor is set to just before
      the output text;
    - when using centered text, the cursor is set to (x,y)

    The rationale is that (when not using centered output) one can do
    several outText()s in a row and always know the correct place without
    explicitly having to compute the width. */
void
outTextF(GfxContext& ctx, int x, int y, int maxw, string_t text)
{
    int width, height;
    GfxFont& fnt = ctx.getFont();

    /* Limit text width */
    while (1) {
        width = fnt.getTextWidth(text);
        if (width <= maxw || text.empty())
            break;

        /* We must remove some characters. Make sure to not hit a
           UTF-8 rune in the middle. */
        string_t::size_type n = text.length() - 1;
        while (n > 0 && Utf8Reader::isContinuationByte(text[n]))
            --n;
        text.erase(n);
    }

    height = fnt.getTextHeight(text);
    GfxPoint align = ctx.getTextAlign();
    int newcy = y;
    y -= height * align.y / 2;

    switch (align.x) {
     case 0:
        /* Left */
        if (!ctx.isTransparentBackground())
            drawBackground(ctx, GfxRect(x, y, maxw, height));
        fnt.outText(ctx, GfxPoint(x, y), text);
        ctx.setCursor(GfxPoint(x + width, newcy));
        break;
     case 1:
        /* Centered */
        {
            int xl = x - maxw/2;
            int xtl = x - width/2;
            if (!ctx.isTransparentBackground())
                drawBackground(ctx, GfxRect(xl, y, maxw, height));
            fnt.outText(ctx, GfxPoint(xtl, y), text);
            ctx.setCursor(GfxPoint(x, newcy));
            break;
        }
     case 2:
        /* Right */
        if (!ctx.isTransparentBackground())
            drawBackground(ctx, GfxRect(x - maxw, y, maxw, height));
        fnt.outText(ctx, GfxPoint(x - width, y), text);
        ctx.setCursor(GfxPoint(x - width, newcy));
        break;
     default:
        ASSERT(0);
    }
}

/** Output Text with fixed maximum width. \overload */
void
outTextF(GfxContext& ctx, int x, int y, int maxw, const char* text)
{
    outTextF(ctx, x, y, maxw, string_t(text));
}

/** Output Text with fixed area.
    See outTextF(GfxContext&,int,int,int,string_t) for more information.
    This fills the area above and below the text as well.

    Note: this does not enforce a maximum height; therefore, the area
    should be at least as tall as the text.

    \param ctx  Context
    \param area Area to fill
    \param text Text to write */
void
outTextF(GfxContext& ctx, const GfxRect& area, string_t text)
{
    if (area.w != 0) {
        GfxFont& fnt = ctx.getFont();

        const int height  = fnt.getTextHeight(text);
        const int align   = ctx.getTextAlign().y;
        const int originY = area.y + area.h * align/2;

        // If filled background is requested, fill.
        if (height < area.h && !ctx.isTransparentBackground()) {
            const int topY = originY - height * align/2;
            if (topY > area.y) {
                drawBackground(ctx, GfxRect(area.x, area.y, area.w, topY - area.y));
            }

            const int botY = topY + height;
            if (botY < area.y + area.h) {
                drawBackground(ctx, GfxRect(area.x, botY, area.w, area.y + area.h - botY));
            }
        }

        // Draw the text
        outTextF(ctx, area.x + ctx.getTextAlign().x * area.w / 2, originY, area.w, text);
    }
}

/** Output Text with fixed area. \overload */
void
outTextF(GfxContext& ctx, const GfxRect& area, const char* text)
{
    outTextF(ctx, area, string_t(text));
}
